# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from reducer import reducer

from ubo_app.load_services import register_service

register_service(
    service_id='notifications',
    label='Notifications',
    reducer=reducer,
)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from typing import Sequence

from redux import BaseAction, BaseEvent, Immutable


class IpAction(BaseAction):
    ...


class IpEvent(BaseEvent):
    ...


class IpUpdateRequestAction(IpAction):
    reset: bool = False


class IpUpdateAction(IpAction):
    interfaces: Sequence[IpNetworkInterface]


class IpUpdateRequestEvent(IpEvent):
    ...


class IpNetworkInterface(Immutable):
    name: str
    ip_addresses: Sequence[str]


class IpState(Immutable):
    interfaces: Sequence[IpNetworkInterface]



"""Application logic."""
from __future__ import annotations

import atexit
import subprocess
from typing import TYPE_CHECKING, Sequence

from debouncer import DebounceOptions, debounce
from headless_kivy_pi.constants import BYTES_PER_PIXEL
from redux import FinishAction, FinishEvent

from ubo_app.store import autorun, dispatch, subscribe_event
from ubo_app.store.main import PowerOffEvent
from ubo_app.store.update_manager import (
    CheckVersionEvent,
    SetUpdateStatusAction,
    UpdateStatus,
    UpdateVersionEvent,
)
from ubo_app.store.update_manager.reducer import ABOUT_MENU_PATH
from ubo_app.store.update_manager.utils import check_version, update
from ubo_app.utils import initialize_board, turn_off_screen, turn_on_screen
from ubo_app.utils.async_ import create_task

if TYPE_CHECKING:
    from ubo_app.menu import MenuApp


def power_off(_: PowerOffEvent) -> None:
    """Power off the device."""
    from headless_kivy_pi import HeadlessWidget

    display = HeadlessWidget._display  # noqa: SLF001
    data = [0] * HeadlessWidget.width * HeadlessWidget.height * BYTES_PER_PIXEL
    display._block(0, 0, HeadlessWidget.width - 1, HeadlessWidget.height - 1, data)  # noqa: SLF001
    subprocess.run(['/usr/bin/env', 'systemctl', 'poweroff', '-i'], check=True)  # noqa: S603

    dispatch(FinishAction())


def setup(app: MenuApp) -> None:
    """Set up the application."""
    turn_on_screen()
    initialize_board()

    subscribe_event(PowerOffEvent, power_off)
    subscribe_event(FinishEvent, lambda *_: app.stop())
    subscribe_event(UpdateVersionEvent, lambda: create_task(update()))
    subscribe_event(CheckVersionEvent, lambda: create_task(check_version()))

    @debounce(
        wait=10,
        options=DebounceOptions(leading=True, trailing=False, time_window=10),
    )
    async def request_check_version() -> None:
        dispatch(SetUpdateStatusAction(status=UpdateStatus.CHECKING))

    @autorun(lambda state: state.main.path)
    async def check_version_caller(
        path: Sequence[str],
        previous_path: Sequence[str],
    ) -> None:
        if path != previous_path and path[:3] == ABOUT_MENU_PATH:
            create_task(request_check_version())

    check_version_caller.subscribe(create_task)
    create_task(request_check_version())

    atexit.register(turn_off_screen)



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

import asyncio
import threading
from typing import TYPE_CHECKING, Coroutine

from redux import FinishEvent

from ubo_app.constants import DEBUG_MODE
from ubo_app.store import subscribe_event

if TYPE_CHECKING:
    from asyncio import Handle


class WorkerThread(threading.Thread):
    def __init__(self: WorkerThread) -> None:
        super().__init__()
        self.loop = asyncio.new_event_loop()
        if DEBUG_MODE:
            self.loop.set_debug(enabled=True)

    def run(self: WorkerThread) -> None:
        asyncio.set_event_loop(self.loop)

        subscribe_event(FinishEvent, self.stop)
        self.loop.run_forever()

    def run_task(self: WorkerThread, task: Coroutine) -> Handle:
        return self.loop.call_soon_threadsafe(self.loop.create_task, task)

    def stop(self: WorkerThread) -> None:
        self.loop.call_soon_threadsafe(self.loop.stop)


thread = WorkerThread()
thread.start()


def _create_task(task: Coroutine) -> Handle:
    return thread.run_task(task)



# ruff: noqa: D100, D101, D102, D103, D104, D105, D107
from __future__ import annotations

import socket
from pathlib import Path

from ubo_app.logging import logger
from ubo_app.store import dispatch
from ubo_app.store.services.rgb_ring import RgbRingSetIsConnectedAction

LM_SOCKET_PATH = Path('/run/ubo').joinpath('ledmanagersocket.sock').as_posix()


class RgbRingClient:
    """The instances of this class send commands to the LED manager.

    The LED Manager runs with root privileges due to hardware DMA security constraints.
    The commands are sent through a socket connection to the LED manager.

    The LED manager is a Python script that runs as a daemon. The LED manager
    is a daemon because it uses DMA to control the LEDs. This is a hardware security
    constraint.

    The LED client is a Python script that runs as a non-root user. The LED client
    function is to serialize LED commands and send them to the LED manager in a secure
    manner.
    """

    def __init__(self: RgbRingClient) -> None:
        self.server_socket: socket.SocketType | None = None
        if Path(LM_SOCKET_PATH).exists():
            self.server_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            try:
                self.server_socket.connect(LM_SOCKET_PATH)
                dispatch(RgbRingSetIsConnectedAction(is_connected=True))
            except Exception as exception:  # noqa: BLE001
                dispatch(RgbRingSetIsConnectedAction(is_connected=False))
                logger.error('Unable to connect to the socket', exc_info=exception)
                return

    def __del__(self: RgbRingClient) -> None:
        if self.server_socket is not None:
            self.server_socket.close()

    def send(self: RgbRingClient, cmd: str) -> None:
        if not self.server_socket:
            return
        self.server_socket.send(cmd.encode('utf-8'))



"""Implement `setup_service` function to set up and enable systemd service."""
from __future__ import annotations

import os
import subprocess
from pathlib import Path
from typing import TypedDict

from ubo_app.constants import INSTALLATION_PATH, USERNAME
from ubo_app.logging import logger


class Service(TypedDict):
    """System service metadata."""

    name: str
    content: str


services: list[Service] = [
    {
        'name': 'ubo-led',
        'content': Path(__file__).parent.joinpath('led.service').open().read(),
    },
    {
        'name': 'ubo-update',
        'content': Path(__file__).parent.joinpath('update.service').open().read(),
    },
    {
        'name': 'ubo-app',
        'content': Path(__file__).parent.joinpath('main.service').open().read(),
    },
]


def bootstrap() -> None:
    """Create the service files and enable the services."""
    for service in services:
        service_file_path = f'/etc/systemd/system/{service["name"]}.service'

        # Ensure we have the required permissions
        if os.geteuid() != 0:
            logger.error('This script needs to be run with root privileges.')
            return

        # Write the service content to the file
        with Path(service_file_path).open('w') as file:
            file.write(
                service['content']
                .replace('{{INSTALLATION_PATH}}', INSTALLATION_PATH)
                .replace('{{USERNAME}}', USERNAME),
            )

        # Enable the service to start on boot
        subprocess.run(
            ['/usr/bin/env', 'systemctl', 'enable', service['name']],  # noqa: S603
            check=True,
        )

        logger.info(
            f"Service '{service['name']}' has been created and enabled.",
        )

    with Path('/etc/polkit-1/rules.d/50-reboot.rules').open('w') as file:
        file.write(
            Path(__file__)
            .parent.joinpath('polkit-reboot.rules')
            .open()
            .read()
            .replace('{{INSTALLATION_PATH}}', INSTALLATION_PATH)
            .replace('{{USERNAME}}', USERNAME),
        )

    subprocess.run(
        [Path(__file__).parent.joinpath('install_wm8960.sh').as_posix()],  # noqa: S603
        check=True,
    )



# ruff: noqa: D100, D101, D102, D103, D104, D107
import os
import sys

from redux import FinishAction

from ubo_app.logging import add_file_handler, add_stdout_handler


def main() -> None:
    """Instantiate the `MenuApp` and run it."""
    add_file_handler()
    add_stdout_handler()

    if len(sys.argv) > 1 and sys.argv[1] == 'bootstrap':
        from ubo_app.system_services.bootstrap import bootstrap

        bootstrap()
        sys.exit(0)

    from headless_kivy_pi import HeadlessWidget

    os.environ['KIVY_METRICS_DENSITY'] = '1'
    os.environ['KIVY_NO_CONFIG'] = '1'
    os.environ['KIVY_NO_FILELOG'] = '1'
    HeadlessWidget.setup_headless({'automatic_fps': True})

    from kivy.clock import Clock

    from ubo_app.load_services import load_services
    from ubo_app.menu import MenuApp

    # Needed since redux is scheduled using Clock scheduler and Clock doesn't run before
    # app is running
    Clock.tick()

    load_services()

    app = MenuApp()

    try:
        app.run()
    finally:
        from ubo_app.store import dispatch

        dispatch(FinishAction())

        # Needed since redux is scheduled using Clock scheduler and Clock doesn't run
        # after app is stopped.
        Clock.tick()


if __name__ == '__main__':
    main()



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from reducer import reducer
from setup import init_service

from ubo_app.load_services import register_service

register_service(
    service_id='wifi',
    label='WiFi',
    reducer=reducer,
)

init_service()



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from dataclasses import replace

from redux import InitAction, InitializationActionError

from ubo_app.store.status_icons import (
    IconState,
    StatusIconsAction,
    StatusIconsRegisterAction,
    StatusIconsState,
)


def reducer(
    state: StatusIconsState | None,
    action: StatusIconsAction | InitAction,
) -> StatusIconsState:
    if state is None:
        if isinstance(action, InitAction):
            return StatusIconsState(icons=[])
        raise InitializationActionError(action)
    if isinstance(action, StatusIconsRegisterAction):
        return replace(
            state,
            icons=sorted(
                [
                    *[
                        icon_status
                        for icon_status in state.icons
                        if icon_status.id != action.id or icon_status.id is None
                    ],
                    IconState(
                        symbol=action.icon,
                        color=action.color,
                        priority=action.priority,
                        id=action.id,
                    ),
                ],
                key=lambda entry: entry.priority,
            ),
        )
    return state



"""Ubo menu application."""
from __future__ import annotations

from ubo_gui.app import UboApp

from .menu_central import MenuAppCentral
from .menu_footer import MenuAppFooter


class MenuApp(MenuAppCentral, MenuAppFooter, UboApp):
    """Menu application."""

    def on_start(self: MenuApp) -> None:
        """Start the application."""
        from ubo_app.side_effects import setup

        setup(self)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from pathlib import Path

from ubo_app.utils.fake import Fake

IS_RPI = Path('/etc/rpi-issue').exists()
if not IS_RPI:
    import sys

    sys.modules['RPi'] = Fake()

from RPi import GPIO  # pyright: ignore [reportMissingImports]  # noqa: E402


def initialize_board() -> None:
    if not IS_RPI:
        return

    GPIO.setmode(GPIO.BCM)
    GPIO.setup(17, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)


def turn_off_screen() -> None:
    if not IS_RPI:
        return

    GPIO.setup(26, GPIO.OUT)
    GPIO.output(26, GPIO.LOW)


def turn_on_screen() -> None:
    if not IS_RPI:
        return

    GPIO.setmode(GPIO.BCM)
    GPIO.setup(26, GPIO.OUT)
    GPIO.output(26, GPIO.HIGH)



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING, Sequence

from kivy.clock import Clock
from redux import FinishAction
from ubo_gui.menu.types import (
    ActionItem,
    ApplicationItem,
    HeadedMenu,
    HeadlessMenu,
    SubMenuItem,
)
from ubo_gui.notification import NotificationWidget

from ubo_app.store import autorun, dispatch
from ubo_app.store.main import PowerOffAction
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationsClearAction,
)
from ubo_app.store.services.sound import SoundPlayChimeAction
from ubo_app.store.update_manager.utils import CURRENT_VERSION, about_menu_items

if TYPE_CHECKING:
    from ubo_gui.menu.types import Item


SETTINGS_MENU = HeadlessMenu(
    title='Settings',
    items=[],
)

APPS_MENU = HeadlessMenu(
    title='Apps',
    items=[],
)


MAIN_MENU = HeadlessMenu(
    title='Main',
    items=[
        SubMenuItem(
            label='Apps',
            icon='apps',
            sub_menu=APPS_MENU,
        ),
        SubMenuItem(
            label='Settings',
            icon='settings',
            sub_menu=SETTINGS_MENU,
        ),
        SubMenuItem(
            label='About',
            icon='info',
            sub_menu=HeadedMenu(
                title='About',
                heading=f'Ubo v{CURRENT_VERSION}',
                sub_heading='A universal dashboard for your Raspberry Pi',
                items=about_menu_items,
            ),
        ),
    ],
)


@autorun(lambda state: state.notifications.unread_count)
def notifications_title(unread_count: int) -> str:
    return f'Notifications ({unread_count})'


@autorun(lambda state: state.notifications.notifications)
def notifications_menu_items(notifications: Sequence[Notification]) -> list[Item]:
    """Return a list of menu items for the notification manager."""

    def notification_widget_builder(
        notification: Notification,
        index: int,
    ) -> type[NotificationWidget]:
        """Return a notification widget for the given notification."""

        class NotificationWrapper(NotificationWidget):
            def __init__(self: NotificationWrapper, **kwargs: object) -> None:
                super().__init__(
                    **kwargs,
                    title=f'Notification ({index+1}/{len(notifications)})',
                    notification_title=notification.title,
                    content=notification.content,
                    color=notification.color,
                    icon=notification.icon,
                )

            def on_dismiss(self: NotificationWrapper) -> None:
                dispatch(
                    NotificationsClearAction(notification=notification),
                )
                Clock.schedule_once(lambda _: self.dispatch('on_close'), -1)

        return NotificationWrapper

    return [
        ApplicationItem(
            label=notification.title,
            icon=notification.icon,
            color='black',
            background_color=notification.color,
            application=notification_widget_builder(notification, index),
        )
        for index, notification in enumerate(notifications)
        if notification.expiry_date is None
        or notification.expiry_date > datetime.now(tz=timezone.utc)
    ]


HOME_MENU = HeadlessMenu(
    title='Dashboard',
    items=[
        SubMenuItem(
            label='',
            sub_menu=MAIN_MENU,
            icon='menu',
            is_short=True,
        ),
        SubMenuItem(
            label='',
            sub_menu=HeadlessMenu(
                title=notifications_title,
                items=notifications_menu_items,
            ),
            color='yellow',
            icon='info',
            is_short=True,
        ),
        ActionItem(
            label='Turn off',
            action=lambda: dispatch(
                SoundPlayChimeAction(name=Chime.FAILURE),
                PowerOffAction(),
                FinishAction(),
            ),
            icon='power_settings_new',
            is_short=True,
        ),
    ],
)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import os
from distutils.util import strtobool

USERNAME = 'ubo'
INSTALLATION_PATH = '/opt/ubo'
DEBUG_MODE = strtobool(os.environ.get('UBO_DEBUG', 'False')) == 1
SERVICES_PATH = (
    os.environ.get('UBO_SERVICES_PATH', '').split(':')
    if os.environ.get('UBO_SERVICES_PATH')
    else []
)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from enum import StrEnum


class GlobalEthernetState(StrEnum):
    CONNECTED = 'Connected'
    DISCONNECTED = 'Disconnected'
    PENDING = 'Pending'
    NEEDS_ATTENTION = 'Needs Attention'
    UNKNOWN = 'Unknown'



# ruff: noqa: D100, D101, D102, D103, D104, D107, PLR2004
from __future__ import annotations

import grp
import logging
import os
import pwd
import socket
import stat
import sys
import time
from pathlib import Path
from threading import Thread
from typing import TYPE_CHECKING, Sequence, cast

import board
import neopixel
from adafruit_blinka.microcontroller.generic_micropython import Pin
from ubo_gui.menu import warnings

from ubo_app.constants import USERNAME

if TYPE_CHECKING:
    from ubo_app.store.services.rgb_ring import Color

BRIGHTNESS = 1.0
NUM_LEDS = 27

LM_SOCKET_PATH = Path(os.getenv('RUNTIME_DIRECTORY', '/run/ubo')).joinpath(
    'ledmanagersocket.sock',
)
# The order of the pixel colors - RGB or GRB.
# Some NeoPixels have red and green reversed!
# For RGBW NeoPixels, simply change the ORDER to RGBW or GRBW.
ORDER = neopixel.GRB

logger = logging.getLogger('led-manager')
logger.setLevel(logging.DEBUG)
logger.debug('Initialising LED-Manager...')


class LEDManager:
    def __init__(self: LEDManager) -> None:
        self.led_ring_present = True
        self.current_color = None
        self.stop: bool = False
        self.current_bright_one = 1
        self.brightness = BRIGHTNESS
        if self.brightness < 0 or self.brightness > 1:
            warnings.warn(
                'Invalid brightness value in config file',
                Warning,
                stacklevel=2,
            )
            self.brightness = 1.0

        self.num_leds = NUM_LEDS
        self.pixels = neopixel.NeoPixel(
            pin=cast(Pin, board.D12),
            n=self.num_leds,
            brightness=1,
            bpp=3,
            auto_write=False,
            pixel_order=ORDER,
        )

    def adjust_brightness(
        self: LEDManager,
        color: Color,
    ) -> Color:
        b = self.brightness
        return (color[0] * b, color[1] * b, color[2] * b)

    def set_brightness(self: LEDManager, b: float) -> None:
        self.brightness = b

    def set_enabled(self: LEDManager, *, enabled: bool = True) -> None:
        if enabled is False:
            self.blank()
        self.led_ring_present = enabled

    def set_all(self: LEDManager, color: Color) -> None:
        if not self.led_ring_present:
            return
        color = self.adjust_brightness(color)
        self.pixels.fill(color)
        self.pixels.show()

    def blank(self: LEDManager) -> None:
        if not self.led_ring_present:
            return
        self.set_all((0, 0, 0))

    def fill_upto(
        self: LEDManager,
        color: Color,
        percentage: float,
        wait: float,
    ) -> None:
        if not self.led_ring_present:
            return
        for i in range(int(self.num_leds * percentage)):
            if self.stop is True:
                self.blank()
                return
            self.pixels[i] = self.adjust_brightness(color)
            time.sleep(wait / 1000)
            self.pixels.show()
        time.sleep(5 * wait / 1000)

    def fill_downfrom(
        self: LEDManager,
        color: Color,
        percentage: float,
        wait: float,
    ) -> None:
        if not self.led_ring_present:
            return
        color = self.adjust_brightness(color)
        self.pixels[: int(self.num_leds * percentage)] = [color] * int(
            self.num_leds * percentage,
        )
        self.pixels.show()
        time.sleep(5 * wait / 1000)
        for i in range(int(self.num_leds * percentage) - 1, -1, -1):
            if self.stop is True:
                self.blank()
                return
            self.pixels[i] = (0, 0, 0)
            time.sleep(wait / 1000)
            self.pixels.show()

    def progress_wheel_step(self: LEDManager, color: Color) -> None:
        if not self.led_ring_present:
            return
        dim_factor = 20
        color = self.adjust_brightness(color)
        self.set_all(
            (color[0] / dim_factor, color[1] / dim_factor, color[2] / dim_factor),
        )
        self.current_bright_one = (self.current_bright_one + 1) % self.num_leds
        before = (self.current_bright_one - 1) % self.num_leds
        if before < 0:
            before = self.num_leds + before
        after = (self.current_bright_one + 1) % self.num_leds

        self.pixels[before] = color
        self.pixels[after] = color
        self.pixels[self.current_bright_one] = color
        self.pixels.show()

    def wheel(self: LEDManager, pos: int) -> Color:
        # Input a value 0 to 255 to get a color value.
        # The colours are a transition r - g - b - back to r.
        if pos < 0 or pos > 0b11111111:
            r = g = b = 0
        elif pos < 0b01010101:
            r = int(pos * 3)
            g = int(0b11111111 - pos * 3)
            b = 0
        elif pos < 0b10101010:
            pos -= 85
            r = int(0b11111111 - pos * 3)
            g = 0
            b = int(pos * 3)
        else:
            pos -= 0b10101010
            r = 0
            g = int(pos * 3)
            b = int(0b11111111 - pos * 3)
        return (r, g, b) if ORDER in (neopixel.RGB, neopixel.GRB) else (r, g, b, 0)

    # wait is in milliseconds
    def rainbow(
        self: LEDManager,
        rounds: int = 50,
        wait: float = 1,
    ) -> None:
        if not self.led_ring_present:
            return
        for _ in range(rounds):
            for j in range(255):
                if self.stop is True:
                    self.blank()
                    return
                for i in range(self.num_leds):
                    pixel_index = (i * 256 // self.num_leds) + j
                    (r, g, b, *_) = self.wheel(pixel_index & 255)
                    self.pixels[i] = (
                        r * self.brightness,
                        g * self.brightness,
                        b * self.brightness,
                    )
                self.pixels.show()
                time.sleep(wait / 1000)
        self.blank()

    def pulse(self: LEDManager, color: Color, wait: float, repetitions: int) -> None:
        # wait is in milliseconds
        # repetitions is the number of retepting pluses
        if not self.led_ring_present:
            return
        dim_steps = 10
        color = self.adjust_brightness(color)
        for _ in range(repetitions):
            for i in range(dim_steps):
                if self.stop is True:
                    self.blank()
                    return
                m = i / dim_steps
                self.pixels.fill((color[0] * m, color[1] * m, color[2] * m))
                self.pixels.show()
                time.sleep(wait / 10000)
            for i in range(1, dim_steps):
                if self.stop is True:
                    self.blank()
                    return
                j = (dim_steps - i) / dim_steps
                self.pixels.fill((color[0] * j, color[1] * j, color[2] * j))
                self.pixels.show()
                time.sleep(wait / 10000)
        self.blank()

    def blink(self: LEDManager, color: Color, wait: float, repetitions: int) -> None:
        # wait is in milliseconds
        # repetitions is the number of blinks
        if not self.led_ring_present:
            return
        color = self.adjust_brightness(color)
        for _ in range(repetitions):
            if self.stop is True:
                self.blank()
                return
            self.pixels.fill((color[0], color[1], color[2]))
            self.pixels.show()
            time.sleep(wait / 1000)
            self.blank()
            time.sleep(1.5 * wait / 1000)

    def spinning_wheel(
        self: LEDManager,
        color: Color,
        wait: float = 1,
        length: int = 5,
        repetitions: int = 5,
    ) -> None:
        if not self.led_ring_present:
            return
        color = self.adjust_brightness(color)
        ring: list[Color] = [(0, 0, 0)] * self.num_leds
        ring[0:length] = [color] * (length)
        if length > self.num_leds:
            warnings.warn(
                f'invalid light strip length! must be under {self.num_leds}',
                Warning,
                stacklevel=2,
            )
            return
        for _ in range(repetitions):
            for i in range(self.num_leds):
                if self.stop is True:
                    self.blank()
                    return
                shifted = ring[i:] + ring[:i]
                # for j in self.num_leds
                self.pixels[:] = shifted
                self.pixels.show()
                time.sleep(wait / 1000)
        self.blank()

    def progress_wheel(self: LEDManager, color: Color, percentage: float) -> None:
        # percentage is a float value between 0 and 1
        if not self.led_ring_present:
            return
        color = self.adjust_brightness(color)
        ring: list[Color] = [(0, 0, 0)] * self.num_leds
        ring[0 : int(self.num_leds * percentage)] = [color] * (
            int(self.num_leds * percentage)
        )
        self.pixels[:] = ring
        self.pixels.show()

    def run_command(self: LEDManager, incoming: Sequence[str]) -> None:  # noqa: C901, PLR0912
        logger.info('---executing command---')
        self.incoming = incoming
        self.stop = False
        if incoming[0] == 'set_enabled' and len(incoming) == 2:
            lm.set_enabled(enabled=incoming[1] == '1')
        # set brightness of LEDs
        if incoming[0] == 'set_brightness' and len(incoming) == 2:
            brightness_value = float(incoming[1])
            if 0 < brightness_value <= 1:
                self.set_brightness(brightness_value)
        if incoming[0] == 'set_all' and len(incoming) == 4:
            lm.set_all((int(incoming[1]), int(incoming[2]), int(incoming[3])))
        if incoming[0] == 'blank':
            self.blank()
        if incoming[0] == 'rainbow' and len(incoming) == 3:
            self.rainbow(int(incoming[1]), float(incoming[2]))
        if incoming[0] == 'pulse' and len(incoming) == 6:
            # pulse(self, color, wait, repetitions):
            self.pulse(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                wait=int(incoming[4]),
                repetitions=int(incoming[5]),
            )
        if incoming[0] == 'blink' and len(incoming) == 6:
            # blink(self, color, wait, repetitions):
            self.blink(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                wait=int(incoming[4]),
                repetitions=int(incoming[5]),
            )
        if incoming[0] == 'progress_wheel_step' and len(incoming) == 4:
            self.progress_wheel_step(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
            )
        if incoming[0] == 'spinning_wheel':  # noqa: SIM102
            # spinning_wheel(self, color, wait, length, repetitions):
            if len(incoming) == 7:
                self.spinning_wheel(
                    (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                    wait=int(incoming[4]),
                    length=int(incoming[5]),
                    repetitions=int(incoming[6]),
                )
        if incoming[0] == 'progress_wheel' and len(incoming) == 5:
            # progress_wheel(self, color, percentage):
            self.progress_wheel(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                percentage=float(incoming[4]),
            )
        if incoming[0] == 'fill_upto' and len(incoming) == 6:
            # fill_upto(self, color, percentage, wait):
            self.fill_upto(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                percentage=float(incoming[4]),
                wait=int(incoming[5]),
            )
        if incoming[0] == 'fill_downfrom' and len(incoming) == 6:
            # fill_upto(self, color, percentage, wait):
            self.fill_downfrom(
                (int(incoming[1]), int(incoming[2]), int(incoming[3])),
                percentage=float(incoming[4]),
                wait=int(incoming[5]),
            )
        self.stop = False


# LED system needs to be root, so need to
# create a socket based command system
# TODO @sassanh: make the socket write permission group based
# this way, apps can be set to be in LED group for permission to control
# the LEDs
if __name__ == '__main__':
    lm = LEDManager()
    lm.incoming = 'spinning_wheel 255 255 255 50 6 100'.split()
    t = Thread(target=lm.run_command, args=(lm.incoming,))
    t.start()

    uid = pwd.getpwnam('root').pw_uid
    gid = grp.getgrnam(USERNAME).gr_gid

    LM_SOCKET_PATH.unlink(missing_ok=True)

    logger.info('LED Manager opening socket...')
    server = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    server.bind(LM_SOCKET_PATH.as_posix())
    permission = (
        stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWGRP | stat.S_IWUSR
    )
    LM_SOCKET_PATH.chmod(permission)
    # this is changing the group to pi
    # it makes sense to make a dedicated rgb-ring group
    # and add pi as user to that group
    os.chown(LM_SOCKET_PATH, uid, gid)

    logger.info('LED Manager Listening...')
    while True:
        try:
            datagram = server.recv(1024)
            if not datagram:
                break
            else:
                incoming_str = datagram.decode('utf-8')
                logger.debug(incoming_str)
                incoming = incoming_str.split()
                # set brightness of LEDs
                if incoming[0] == 'set_brightness':
                    if len(incoming) == 2:
                        brightness_value = float(incoming[1])
                        if 0 < brightness_value <= 1:
                            lm.set_brightness(brightness_value)
                else:
                    lm.stop = True
                    while t.is_alive():
                        time.sleep(0.1)
                    # save some data before overriding the object
                    logger.debug('---starting new led thread--')
                    t = Thread(target=lm.run_command, args=(incoming,))
                    t.start()
        except KeyboardInterrupt:
            logger.debug('Interrupted')
            server.close()
            try:
                sys.exit(0)
            except SystemExit:
                os._exit(0)
    logger.info('Shutting down...')
    server.close()
    LM_SOCKET_PATH.unlink()



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from typing import Callable

from kivy.base import Clock
from redux import (
    BaseCombineReducerState,
    CombineReducerAction,
    CreateStoreOptions,
    combine_reducers,
    create_store,
)

from ubo_app.logging import logger
from ubo_app.store.main import MainAction, MainState
from ubo_app.store.main.reducer import reducer as main_reducer
from ubo_app.store.services.camera import CameraAction, CameraEvent
from ubo_app.store.services.ip import IpAction, IpEvent, IpState
from ubo_app.store.services.keypad import KeypadEvent
from ubo_app.store.services.notifications import NotificationsAction, NotificationsState
from ubo_app.store.services.sound import SoundAction, SoundState
from ubo_app.store.services.wifi import WiFiAction, WiFiEvent, WiFiState
from ubo_app.store.status_icons import StatusIconsAction, StatusIconsState
from ubo_app.store.status_icons.reducer import reducer as status_icons_reducer
from ubo_app.store.update_manager import UpdateManagerAction, UpdateManagerState
from ubo_app.store.update_manager.reducer import reducer as update_manager_reducer


class RootState(BaseCombineReducerState):
    main: MainState
    status_icons: StatusIconsState
    update_manager: UpdateManagerState
    sound: SoundState
    wifi: WiFiState
    ip: IpState
    notifications: NotificationsState


ActionType = (
    CombineReducerAction
    | StatusIconsAction
    | UpdateManagerAction
    | MainAction
    | SoundAction
    | CameraAction
    | WiFiAction
    | IpAction
    | NotificationsAction
)
EventType = KeypadEvent | CameraEvent | WiFiEvent | IpEvent

root_reducer, root_reducer_id = combine_reducers(
    state_type=RootState,
    action_type=ActionType,
    event_type=EventType,
    main=main_reducer,
    status_icons=status_icons_reducer,
    update_manager=update_manager_reducer,
)


def scheduler(callback: Callable[[], None], *, interval: bool) -> None:
    Clock.create_trigger(lambda _: callback(), 0, interval=interval)()


store = create_store(
    root_reducer,
    CreateStoreOptions(
        auto_init=True,
        scheduler=scheduler,
        action_middleware=lambda action: logger.debug(
            'Action dispatched',
            extra={'action': action},
        ),
        event_middleware=lambda event: logger.debug(
            'Event dispatched',
            extra={'event': event},
        ),
    ),
)

autorun = store.autorun
dispatch = store.dispatch
subscribe = store.subscribe
subscribe_event = store.subscribe_event

__ALL__ = (autorun, dispatch, subscribe, subscribe_event)



# pyright: reportMissingImports=false,reportMissingModuleSource=false
"""Module for managing audio playback and recording."""
from __future__ import annotations

import contextlib
import math
import time
import wave

from headless_kivy_pi import IS_RPI

from ubo_app.logging import logger

if not IS_RPI:
    import sys

    from ubo_app.utils.fake import Fake

    sys.modules['alsaaudio'] = Fake()
    sys.modules['pulsectl'] = Fake()
    sys.modules['pyaudio'] = Fake()

import alsaaudio
import pulsectl
import pyaudio

# TODO(@sassanh): It's a dynamic value, so we should probably get it from somewhere
RESPEAKER_INDEX = 1
CHUNK_SIZE = 1024


def set_default_sink(*, name: str) -> None:
    """Set the default sink to the sink with the given name."""
    with pulsectl.Pulse('set-default-sink-by-description') as pulse:
        for sink in pulse.sink_list():
            if sink.proplist['alsa.long_card_name'] == name:
                pulse.sink_default_set(sink)
                logger.info('Set default sink to', extra={'sink': sink})
                return
        logger.error('No sink found', extra={'name_': name})


def linear_to_logarithmic(volume_linear: float) -> int:
    """Convert a linear volume to a logarithmic volume.

    Assuming volume_linear is between 0 and 1
    Convert it to a logarithmic scale.
    """
    if volume_linear == 0:
        return 0
    return round(100 * math.log(volume_linear) / math.log(100))


class AudioManager:
    """Class for managing audio playback and recording."""

    stream: pyaudio.PyAudio.Stream | None = None

    def __init__(self: AudioManager) -> None:
        """Initialize the audio manager."""
        # create an audio object
        self.pyaudio = pyaudio.PyAudio()
        self.is_playing = False
        self.should_stop = False

        cardindex = None
        try:
            cards = alsaaudio.cards()
            cardindex = cards.index(
                next(card for card in cards if 'wm8960' in card),
            )
        except StopIteration:
            logger.error('No audio card found')

        if cardindex is None:
            return

        mixer = alsaaudio.Mixer(
            control='Right Output Mixer PCM',
            cardindex=cardindex,
        )
        mixer.setmute(0)
        mixer = alsaaudio.Mixer(
            control='Left Output Mixer PCM',
            cardindex=cardindex,
        )
        mixer.setmute(0)

    def __del__(self: AudioManager) -> None:
        """Clean up the audio manager."""
        self.close_stream()
        self.pyaudio.terminate()

    def close_stream(self: AudioManager) -> None:
        """Clean up the audio manager."""
        if self.stream:
            self.should_stop = True
            while self.is_playing:
                time.sleep(0.05)
            self.should_stop = False
            with contextlib.suppress(Exception):
                self.stream.close()
            self.stream = None

    def play(self: AudioManager, filename: str) -> None:
        """Play a waveform audio file.

        Parameters
        ----------
        filename : str
            Path to wav file
        """
        # open the file for reading.
        self.close_stream()

        logger.info('Opening audio file for playback', extra={'filename_': filename})
        try:
            with wave.open(filename, 'rb') as wf:
                self.is_playing = True
                stream = self.stream = self.pyaudio.open(
                    format=self.pyaudio.get_format_from_width(wf.getsampwidth()),
                    channels=wf.getnchannels(),
                    rate=wf.getframerate(),
                    output=True,
                    output_device_index=RESPEAKER_INDEX,
                )
                data = wf.readframes(CHUNK_SIZE)
                while data and not self.should_stop and stream.is_active():
                    stream.write(data)
                    data = wf.readframes(CHUNK_SIZE)
        except Exception as exception:  # noqa: BLE001
            logger.error(
                'Something went wrong while playing an audio file',
                exc_info=exception,
            )
        finally:
            self.is_playing = False
            self.close_stream()

    def set_playback_mute(self: AudioManager, *, mute: bool = False) -> None:
        """Set the playback mute of the audio output.

        Parameters
        ----------
        mute : bool
            Mute to set
        """
        mixer = alsaaudio.Mixer(control='Master')
        mixer.setmute(1 if mute else 0)

    def set_playback_volume(self: AudioManager, volume: float = 0.8) -> None:
        """Set the playback volume of the audio output.

        Parameters
        ----------
        volume : float
            Volume to set, a float between 0 and 1
        """
        if volume < 0 or volume > 1:
            msg = 'Volume must be between 0 and 1'
            raise ValueError(msg)
        mixer = alsaaudio.Mixer(control='Master')
        mixer.setvolume(round(volume * 100))

    def set_capture_volume(self: AudioManager, volume: float = 0.8) -> None:
        """Set the capture volume of the audio output.

        Parameters
        ----------
        volume : float
            Volume to set, a float between 0 and 1
        """
        if volume < 0 or volume > 1:
            msg = 'Volume must be between 0 and 1'
            raise ValueError(msg)
        mixer = alsaaudio.Mixer(control='Capture')
        mixer.setrec(round(volume * 100))



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import atexit
import logging
import sys
from typing import Mapping, cast

from ubo_app.constants import DEBUG_MODE

VERBOSE = 5


class UboLogger(logging.getLoggerClass()):
    def __init__(self: UboLogger, name: str, level: int = logging.NOTSET) -> None:
        super().__init__(name, level)

        logging.addLevelName(VERBOSE, 'VERBOSE')

    def verbose(  # noqa: PLR0913
        self: UboLogger,
        msg: object,
        *args: tuple[object],
        exc_info: logging._ExcInfoType | None = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Mapping[str, object] | None = None,
    ) -> None:
        if self.isEnabledFor(VERBOSE):
            self._log(
                VERBOSE,
                msg,
                *args,
                exc_info=exc_info,
                stack_info=stack_info,
                stacklevel=stacklevel,
                extra=extra,
            )


logging.setLoggerClass(UboLogger)

logger = cast(UboLogger, logging.getLogger('ubo-app'))
if DEBUG_MODE:
    logger.setLevel(logging.DEBUG)
else:
    logger.setLevel(logging.INFO)
logger.propagate = False


class ExtraFormatter(logging.Formatter):
    def_keys = (
        'name',
        'msg',
        'args',
        'levelname',
        'levelno',
        'pathname',
        'filename',
        'module',
        'exc_info',
        'exc_text',
        'stack_info',
        'lineno',
        'funcName',
        'created',
        'msecs',
        'relativeCreated',
        'thread',
        'threadName',
        'processName',
        'process',
        'message',
    )

    def format(self: ExtraFormatter, record: logging.LogRecord) -> str:
        string = super().format(record)
        extra = {k: v for k, v in record.__dict__.items() if k not in self.def_keys}
        if len(extra) > 0:
            string += ' - extra: ' + str(extra)

        return string


def add_stdout_handler() -> None:
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)
    stdout_handler.setFormatter(
        ExtraFormatter(
            '%(created)f [%(levelname)s] %(message)s',
            '%Y-%m-%d %H:%M:%S',
        ),
    )
    logger.addHandler(stdout_handler)

    atexit.register(stdout_handler.flush)


def add_file_handler() -> None:
    file_handler = logging.FileHandler('ubo-app.log')
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(
        ExtraFormatter(
            '%(created)f [%(levelname)s] %(message)s',
            '%Y-%m-%d %H:%M:%S',
        ),
    )
    logger.addHandler(file_handler)

    atexit.register(file_handler.flush)


__all__ = ('logger', 'add_stdout_handler', 'add_file_handler')



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from dataclasses import field
from typing import TYPE_CHECKING, Sequence

from redux import BaseAction, BaseEvent, FinishAction, Immutable, InitAction

from ubo_app.store.services.keypad import KeypadAction
from ubo_app.store.status_icons import StatusIconsAction

if TYPE_CHECKING:
    from typing_extensions import TypeAlias
    from ubo_gui.menu.types import Item, Menu


class RegisterAppAction(BaseAction):
    menu_item: Item


class RegisterRegularAppAction(RegisterAppAction):
    ...


class RegisterSettingAppAction(RegisterAppAction):
    ...


class PowerOffAction(BaseAction):
    ...


class PowerOffEvent(BaseEvent):
    ...


class MainState(Immutable):
    menu: Menu | None = None
    path: Sequence[str] = field(default_factory=list)


class InitEvent(BaseEvent):
    ...


class SetMenuPathAction(BaseAction):
    path: Sequence[str]


MainAction: TypeAlias = (
    InitAction
    | FinishAction
    | PowerOffAction
    | StatusIconsAction
    | KeypadAction
    | RegisterAppAction
    | SetMenuPathAction
)



# ruff: noqa: D100, D101, D102, D103, D104, D105, D107
from __future__ import annotations

import asyncio
import importlib
import importlib.abc
import importlib.util
import os
import sys
import threading
import traceback
import uuid
from importlib.machinery import PathFinder, SourceFileLoader
from pathlib import Path
from typing import TYPE_CHECKING, Any, Sequence, cast

from redux import CombineReducerRegisterAction, FinishEvent, ReducerType

from ubo_app.constants import DEBUG_MODE, SERVICES_PATH
from ubo_app.logging import logger

if TYPE_CHECKING:
    from importlib.machinery import ModuleSpec
    from types import ModuleType

ROOT_PATH = Path(__file__).parent
REGISTERED_PATHS: dict[Path, UboServiceThread] = {}


# Customized module finder and module loader for ubo services to avoid mistakenly
# loading conflicting names in different services.
# This is a temporary hack until each service runs in its own process.
class UboServiceModuleLoader(SourceFileLoader):
    def get_filename(self: UboServiceModuleLoader, name: str | None = None) -> str:
        return super().get_filename(name.split(':')[1] if name else name)

    def create_module(
        self: UboServiceModuleLoader,
        spec: ModuleSpec,
    ) -> ModuleType | None:
        if spec.name in sys.modules:
            return sys.modules[spec.name]
        return super().create_module(spec)


class UboServiceLoopLoader(importlib.abc.Loader):
    def __init__(self: UboServiceLoopLoader, thread: UboServiceThread) -> None:
        self.thread = thread

    def exec_module(self: UboServiceLoopLoader, module: ModuleType) -> None:
        cast(Any, module)._create_task = (  # noqa: SLF001
            lambda task: self.thread.loop.call_soon_threadsafe(
                self.thread.loop.create_task,
                task,
            )
        )

    def __repr__(self: UboServiceLoopLoader) -> str:
        return f'{self.thread.path}'


class UboServiceFinder(importlib.abc.MetaPathFinder):
    def find_spec(
        self: UboServiceFinder,
        fullname: str,
        _: Sequence[str] | None,
        target: ModuleType | None = None,
    ) -> ModuleSpec | None:
        stack = traceback.extract_stack()
        matching_path = next(
            (
                registered_path
                for stack_path in stack[::-1]
                for registered_path in REGISTERED_PATHS
                if stack_path.filename.startswith(registered_path.as_posix())
            ),
            None,
        )

        if matching_path:
            thread = REGISTERED_PATHS[matching_path]
            module_name = f'{thread.service_id}:{fullname}'

            if fullname == 'ubo_app.utils.loop':
                return importlib.util.spec_from_loader(
                    module_name,
                    UboServiceLoopLoader(thread),
                )

            spec = PathFinder.find_spec(
                fullname,
                [matching_path.as_posix()],
                target,
            )
            if spec and spec.origin:
                spec.name = module_name
                spec.loader = UboServiceModuleLoader(fullname, spec.origin)
            return spec
        return None


sys.meta_path.insert(0, UboServiceFinder())


class UboServiceThread(threading.Thread):
    def __init__(
        self: UboServiceThread,
        path: Path,
        service_id: str,
    ) -> None:
        super().__init__()
        self.service_id = service_id
        self.path = path
        self.loop = asyncio.new_event_loop()
        self.module = None
        if DEBUG_MODE:
            self.loop.set_debug(enabled=True)
        try:
            if path.exists():
                module_name = f'{service_id}:ubo_handle'
                self.spec = PathFinder.find_spec(
                    'ubo_handle',
                    [path.as_posix()],
                    None,
                )
                if not self.spec or not self.spec.origin:
                    return
                self.spec.name = module_name
                self.spec.loader = UboServiceModuleLoader(
                    'ubo_handle',
                    self.spec.origin,
                )
                self.module = importlib.util.module_from_spec(self.spec)
                sys.modules[module_name] = self.module

        except Exception as exception:  # noqa: BLE001
            logger.error(f'Error loading "{path}"', exc_info=exception)

    def run(self: UboServiceThread) -> None:
        from ubo_app import store

        asyncio.set_event_loop(self.loop)
        if self.module and self.spec and self.spec.loader:
            try:
                self.spec.loader.exec_module(self.module)
            except Exception as exception:  # noqa: BLE001
                logger.error(f'Error loading "{self.path}"', exc_info=exception)
                return

        store.subscribe_event(FinishEvent, self.stop)
        self.loop.run_forever()

    def stop(self: UboServiceThread) -> None:
        self.loop.call_soon_threadsafe(self.loop.stop)


def register_service(
    service_id: str,
    label: str,
    reducer: ReducerType | None = None,
) -> None:
    from ubo_app import store

    logger.info(
        'Registering ubo serivce',
        extra={
            'service_id': service_id,
            'label': label,
            'has_reducer': reducer is not None,
        },
    )

    if reducer is not None:
        store.dispatch(
            CombineReducerRegisterAction(
                _id=store.root_reducer_id,
                key=service_id,
                reducer=reducer,
            ),
        )


def load_services() -> None:
    for services_directory_path in [
        ROOT_PATH.joinpath('services').as_posix(),
        *SERVICES_PATH,
    ]:
        if Path(services_directory_path).is_dir():
            for service_path in Path(services_directory_path).iterdir():
                service_id = uuid.uuid4().hex
                if not service_path.is_dir():
                    continue
                current_path = Path().absolute()
                os.chdir(service_path.as_posix())

                thread = UboServiceThread(
                    service_path,
                    service_id,
                )
                REGISTERED_PATHS[service_path] = thread
                thread.start()

                os.chdir(current_path)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from dataclasses import replace
from typing import Sequence, cast

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)
from ubo_gui.menu.types import Menu, SubMenuItem

from ubo_app.store.main import (
    InitEvent,
    MainAction,
    MainState,
    PowerOffAction,
    PowerOffEvent,
    RegisterAppAction,
    RegisterRegularAppAction,
    SetMenuPathAction,
)
from ubo_app.store.services.keypad import (
    Key,
    KeypadEvent,
    KeypadKeyPressAction,
    KeypadKeyPressEvent,
)
from ubo_app.store.services.sound import SoundChangeVolumeAction, SoundDevice


def reducer(  # noqa: C901
    state: MainState | None,
    action: MainAction,
) -> ReducerResult[
    MainState,
    SoundChangeVolumeAction,
    KeypadEvent | InitEvent | PowerOffEvent,
]:
    if state is None:
        if isinstance(action, InitAction):
            from ._menus import HOME_MENU

            return CompleteReducerResult(
                state=MainState(menu=HOME_MENU),
                events=[InitEvent()],
            )
        raise InitializationActionError(action)

    if isinstance(action, KeypadKeyPressAction):
        actions: list[SoundChangeVolumeAction] = []
        if action.key == Key.UP and len(state.path) == 1:
            actions.append(
                SoundChangeVolumeAction(
                    amount=0.05,
                    device=SoundDevice.OUTPUT,
                ),
            )
        if action.key == Key.DOWN and len(state.path) == 1:
            actions.append(
                SoundChangeVolumeAction(
                    amount=-0.05,
                    device=SoundDevice.OUTPUT,
                ),
            )
        return CompleteReducerResult(
            state=state,
            actions=actions,
            events=[
                KeypadKeyPressEvent(key=action.key),
            ],
        )

    if isinstance(action, RegisterAppAction):
        from ubo_gui.menu import Item, menu_items

        menu = state.menu
        parent_index = 0 if isinstance(action, RegisterRegularAppAction) else 1

        if not menu:
            return state

        root_menu_items = menu_items(menu)

        main_menu_item: Item = root_menu_items[0]
        if not isinstance(main_menu_item, SubMenuItem):
            msg = 'Main menu item is not a `SubMenuItem`'
            raise TypeError(msg)

        main_menu_items = menu_items(cast(Menu, main_menu_item.sub_menu))

        desired_menu_item = main_menu_items[parent_index]
        if not isinstance(desired_menu_item, SubMenuItem):
            menu_title = (
                'Applications'
                if isinstance(action, RegisterRegularAppAction)
                else 'Settings'
            )
            msg = f'{menu_title} menu item is not a `SubMenuItem`'
            raise TypeError(msg)

        new_items = [
            *cast(Sequence[Item], cast(Menu, desired_menu_item.sub_menu).items),
            action.menu_item,
        ]
        desired_menu_item = replace(
            desired_menu_item,
            sub_menu=replace(
                cast(Menu, desired_menu_item.sub_menu),
                items=new_items,
            ),
        )
        main_menu_item = replace(
            main_menu_item,
            sub_menu=replace(
                cast(Menu, main_menu_item.sub_menu),
                items=[
                    desired_menu_item if index == parent_index else item
                    for index, item in enumerate(main_menu_items)
                ],
            ),
        )

        return replace(
            state,
            menu=replace(
                menu,
                items=[
                    main_menu_item if index == 0 else item
                    for index, item in enumerate(root_menu_items)
                ],
            ),
        )

    if isinstance(action, SetMenuPathAction):
        return replace(state, path=action.path)

    if isinstance(action, PowerOffAction):
        return CompleteReducerResult(
            state=state,
            events=[PowerOffEvent()],
        )

    return state
